<?php

namespace Extensions\ZeroRatedMobileAccess;

use ApiBase;
use ApiFormatBase;
use ApiMain;
use FauxRequest;
use FormatJson;

/**
 * Utility class to avoid any warnings or any other additions by the API framework
 * Class ApiZeroConfigPrinter
 * @package Extensions\ZeroRatedMobileAccess
 */
class ApiZeroConfigPrinter extends \ApiFormatJson {
	private $moduleName;

	public function __construct( $main, $format, $moduleName ) {
		parent::__construct( $main, $format );
		$this->moduleName = $moduleName;
	}

	public function execute() {
		// Make sure we only output the data we
		$res = $this->getResult();
		$data = $res->getData();
		$res->reset();
		// @todo: add ApiResult::NO_SIZE_CHECK when core update is live
		if ( array_key_exists( $this->moduleName, $data ) ) {
			foreach ( $data[$this->moduleName] as $k => $v ) {
				$res->addValue( null, $k, $v );
			}
		}
		parent::execute();
	}
}

class ApiZeroConfig extends ApiBase {
	/**
	 * Override built-in handling of format parameter.
	 * Only JSON is supported.
	 * @return ApiFormatBase
	 */
	public function getCustomPrinter() {
		return new ApiZeroConfigPrinter( $this->getMain(), 'json', $this->getModuleName() );
	}

	public function execute() {
		$this->getMain()->setCacheMaxAge( 1 * 60 ); // seconds
		$this->getMain()->setCacheMode( 'public' );
		$result = $this->getResult();

		$params = $this->extractRequestParams();
		$type = $params['type'];
		$moduleName = $this->getModuleName();

		switch ( $type ) {
			case 'config':
				$state = PageRenderingHooks::getState( $this );
				$config = $state->getConfig();
				if ( $config !== null ) {
					$result->addValue( $moduleName, 'enabled', true );
					$result->addValue( $moduleName, 'id', $state->getConfigId() );
					$result->addValue( $moduleName, 'showImages', $config['showImages'] );
					$result->addValue( $moduleName, 'whitelistedLangs', $config['whitelistedLangs'] );
					$result->addValue( $moduleName, 'enableHttps', $config['enableHttps'] );
					$result->addValue( $moduleName, 'sites', $config['sites'] );
				} else {
					$result->addValue( $moduleName, 'enabled', false );
				}
				$out = $result->getOutput();
				$out->addVaryHeader( 'X-CS' );
				$out->addVaryHeader( 'X-Subdomain' );
				$out->addVaryHeader( 'X-Forwarded-By' );
				break;

			case 'message':
				if ( ! ( isset( $params['agent'] ) && strlen( $params['agent'] ) ) ) {
					$this->dieUsage( 'Non-empty agent parameter is required when type param is message.',
						'missing_valid_required_parameter', 500 );
				}
				$state = PageRenderingHooks::getState( $this );
				$config = $state->getConfig();
				if ( $config !== null && !$config['disableApps'] ) {
					$result->addValue( $moduleName, 'message', $state->getBannerText( $config, null, null ) );
				} else {
					// json_encode will ultimately return as an empty array
				}
				$out = $result->getOutput();
				$out->addVaryHeader( 'X-CS' );
				$out->addVaryHeader( 'X-Subdomain' );
				$out->addVaryHeader( 'X-Forwarded-By' );
				break;

			case 'proxies':
			case 'carriers':
				global $wgZeroRatedMobileAccessEnableZeroConfigPages;
				if ( $wgZeroRatedMobileAccessEnableZeroConfigPages !== true ) {
					$this->dieUsage( 'Zero config is not enabled for this site', 'data', 500 );
				}
				$result->disableSizeCheck();

				$reqParamsOriginal = array(
					'action' => 'query',
					'generator' => 'allpages',
					'gaplimit' => 'max',
					'gapnamespace' => '480',
					'prop' => 'revisions',
					'rvprop' => 'content',
					'continue' => '',
				);
				$isProxies = $type === 'proxies';
				if ( $isProxies ) {
					$reqParamsOriginal['gapprefix'] = '-';
				} elseif ( $type === 'carriers' ) {
					// All carriers are in NNN-NN format, which means we can safely enumerate
					// begining from 0 to avoid proxies that start with a '-'.
					$reqParamsOriginal['gapfrom'] = '0';
				}
				$reqParams = $reqParamsOriginal;

				while ( true ) {
					// Get needed data
					$req = new FauxRequest( $reqParams );
					$api = new ApiMain( $req );
					$api->execute();
					$data = $api->getResultData();
					// @bug: errors are not returned, they are thrown, when calling api with faux request
					if ( array_key_exists( 'error', $data ) ) {
						$this->dieUsage( 'Error getting data', 'data', 500, $data['error'] );
					}
					foreach ( (array)$data['query']['pages'] as $pageId => $page ) {
						$rawText = $page['revisions'][0]['*'];
						$json = FormatJson::decode( $rawText, true ) ?: false;
						if ( !is_array( $json ) ) {
							$this->dieUsage( "Unable to parse json of page $pageId", 'json', 500 );
						}
						if ( !array_key_exists( 'ips', $json ) ) {
							continue;
						}
						$ips = $json['ips'];
						if ( !is_array( $ips ) || count( $ips ) === 0 ) {
							continue;
						}
						$title = str_replace( 'Zero:', '', $page['title'] );
						if ( $isProxies ) {
							$title = ucfirst( strtolower( substr( $title, 1 ) ) ); // remove '-', and make it "Opera"
						}
						$result->addValue( $moduleName, $title, $ips );
					}
					if ( !array_key_exists( 'continue', $data ) ) {
						break;
					}
					$reqParams = array_merge( $reqParamsOriginal, $data['continue'] );
				}
				break;
		}
	}

	public function getAllowedParams() {
		return array(
			'type' => array(
				ApiBase::PARAM_DFLT => 'config',
				ApiBase::PARAM_TYPE => array(
					'config',
					'proxies',
					'carriers',
					'message',
				)
			),
			'agent' => array(
				ApiBase::PARAM_DFLT => '',
			),
		);
	}

	public function getParamDescription() {
		return array(
			'type' => array(
				'What kind of Zero info is needed',
				'  config   - get a trimmed-down carrier configuration based on the X-CS header',
				'    These types only work on the site that has zero config enabled (e.g. meta):',
				'  proxies  - list of all ips of the proxies (pages begin with "-")',
				'  carriers - list of all ips of all carriers',
				'  message  - get lang-appropriate verbiage based on X-CS header; requires agent param',
			),
			'agent' => array(
				'When setting type to message, include a source agent value as well',
			),
		);
	}

	public function getDescription() {
		return 'Get configuration of the Zero extension';
	}

	public function getExamples() {
		return array(
			'api.php?action=zeroconfig&type=config',
			'api.php?action=zeroconfig&type=message&agent=neatapp',
		);
	}
}
