<?php

namespace Extensions\ZeroRatedMobileAccess;
use FormatJson;
use ObjectCache;
use Message;

/**
 * JSON Zero Config
 *
 * @file
 * @ingroup Extensions
 * @ingroup ZeroRatedMobileAccess
 *
 * @author Yuri Astrakhan <yurik@wikimedia.org>
 */
class CarrierConfig {

	private $rawText;
	private $json, $originalJson;
	private $errorMessage = null;
	private $errors = array();
	private $config = array();
	private $defaultUsed = array();
	private $sameAsDefault = array();
	private $unknownFields;
	private $isSaving;

	/**
	 * @param $rawText
	 * @param bool $isSaving if true, performs extra validation checks the ZeroConfig page editing
	 */
	function __construct( $rawText, $isSaving = false ) {
		$this->rawText = $rawText;
		$this->isSaving = $isSaving;
		$this->validate();
	}

	public function isValidJson() {
		return $this->originalJson !== null;
	}

	public function getOriginalJson() {
		return $this->originalJson;
	}

	/**
	 * This configuration has errors
	 * @return bool
	 */
	public function isError() {
		return $this->errorMessage !== null;
	}

	/**
	 * Returns null if no error, or the rendering of error messages
	 * @return string|null
	 */
	public function getErrorMessage() {
		$err = $this->errorMessage;
		if ( $err !== null ) {
			foreach ( $this->errors as $f => $e ) {
				$err .= '<br /> *' . wfMessage( 'zeroconfig-field_error', $f, $e );
			}
		}
		return $err;
	}

	/**
	 * Returns null if no error, or the rendering of error messages
	 * @param \Status $status
	 */
	public function setErrorMessageStatus( $status ) {
		$err = $this->errorMessage;
		if ( $err !== null ) {
			$status->error( $err );
			foreach ( $this->errors as $f => $e ) {
				$status->error( wfMessage( 'zeroconfig-field_error', $f, $e ) );
			}
			$status->ok = false;
		}
	}

	/**
	 * Get an array of fields that were not present in the configuration
	 * @return array
	 */
	public function getDefaultUsed() {
		return $this->defaultUsed;
	}

	/**
	 * Get an array of fields that have the same value as default
	 * @return array
	 */
	public function getSameAsDefault() {
		return $this->sameAsDefault;
	}

	/**
	 * Get a dictionary of all unrecognized fields with their values
	 * @return Array
	 */
	public function getUnknownFields() {
		return $this->unknownFields;
	}

	/**
	 * Get configuration data
	 * @param bool $withDefaults If true, the result will include default values
	 * @return array
	 */
	public function getData( $withDefaults = true ) {
		if ( $withDefaults ) {
			return $this->config;
		} else {
			$tmp = $this->config;
			foreach ( $this->defaultUsed as $fld => $val ) {
				unset( $tmp[$fld] );
			}
			return $tmp;
		}
	}

	/**
	 * Ensure that current config contains user in the 'admins' section
	 * @param \User $user
	 */
	public function addAdmin( $user ) {
		$admins = $this->config['admins'];
		$name = \User::getCanonicalName( $user->getName(), 'usable' );
		if ( $name  ) {
			$admins[] = $name;
			$this->config['admins'] = self::normalizeAdmins( $admins );
		}
	}

	private function validate() {

		$json = FormatJson::decode( $this->rawText, true ) ?: false;
		if ( !is_array( $json ) ) {
			if ( $this->isSaving ) {
				// The most common error is the trailing comma in a list. Attempt to remove it.
				$json = preg_replace( '/,[ \t]*([\r\n]*[ \t]*[\]}])/', '$1', $this->rawText );
				$json = FormatJson::decode( $json, true ) ?: false;
			}
			if ( !is_array( $json ) ) {
				$this->errorMessage = wfMessage( 'zeroconfig-json' );
				return;
			}
		}
		$this->json = $json;
		$this->originalJson = $json;

		$validateBool = function( $v ) {
			return is_bool( $v ) ? null : wfMessage( 'zeroconfig-opt_bool' );
		};
		$validateStr = function( $v ) {
			return is_string( $v ) ? null : wfMessage( 'zeroconfig-opt_string' );
		};

		// Need a local var to be able to use it in lambda closures
		$isSaving = $this->isSaving;

		// Optional comment
		$this->check( 'comment', '', $validateStr );

		//'enabled' => true,          // Config is enabled
		$this->check( 'enabled', true, $validateBool );

		// List of additional partner admins for this entry
		if ( $this->isSaving ) {
			$this->check( 'admins', array(),
				function( $v ) {
					if ( is_string( $v ) ) {
						$v = array( $v );
					} elseif ( !CarrierConfig::isArray( $v, false ) || !CarrierConfig::isArrayOfStrings( $v ) ) {
						return wfMessage( 'zeroconfig-admins' );
					}
					$v2 = array();
					foreach ( $v as $name ) {
						$usr = \User::newFromName( $name );
						if ( $usr === false || $usr->getId() === 0 ) {
							return wfMessage( 'zeroconfig-admins' );
						}
						$v2[] = $usr->getName();
					}
					return CarrierConfig::normalizeAdmins( $v2 );
				} );
		}

		// Which sites are whitelisted. Default - both m & zero wiki
		$this->check( 'sites',
			array(
				'm.wikipedia',
				'zero.wikipedia',
			),
			function( $v ) {
				if ( is_string( $v ) ) {
					$oldValues = array(
						'zero' => 'zero.wikipedia',
						'm' => 'm.wikipedia',
						'both' => array( 'm.wikipedia', 'zero.wikipedia' ) );
					$v = strtolower( $v );
					if ( array_key_exists( $v, $oldValues ) ) {
						$v = $oldValues[$v];
					}
					if ( is_string( $v ) ) {
						$v = array( $v );
					}
				}
				$validValues = array(
					'm.wikipedia',
					'zero.wikipedia',
				);
				if ( is_array( $v ) ) {
					$v = array_map( 'strtolower', $v );
					// FIXME: remove this after refreshing all meta configs
					foreach ( $v as &$item ) {
						if ( $item === 'zero.wiki' ) {
							$item = 'zero.wikipedia';
						} elseif ( $item === 'm.wiki' ) {
							$item = 'm.wikipedia';
						}
					}
					if ( count( array_intersect( $v, $validValues ) ) !== count( $v ) ) {
						$v = false;
					} else {
						$v = array_unique( $v );
						sort( $v );
					}
				}
				if ( CarrierConfig::isArray( $v, false )
					&& count( $v ) > 0
					&& CarrierConfig::isArrayOfStrings( $v )
				) {
					return $v;
				}
				return wfMessage( 'zeroconfig-sites', "'" . implode( "', '", $validValues ) . "'" )
					->numParams( count( $validValues ) );
			} );

		// If carrier wants to suppress zero messaging in apps
		$this->check( 'disableApps', false, $validateBool );

		//'name' => null,             // Map of localized partner names
		$this->check( 'name', null,
			function( $v ) {
				return CarrierConfig::isArray( $v, true )
					&& CarrierConfig::isArrayOfLangs( array_keys( $v ) )
					&& CarrierConfig::isArrayOfStrings( $v )
					? CarrierConfig::sortArray( $v ) : wfMessage( 'zeroconfig-name' );
			} );

		//'banner' => null,           // Map of localized banner texts with {{PARTNER}} placeholder
		$this->check( 'banner', array(),
			function( $v ) {
				return CarrierConfig::isArray( $v, true )
					&& CarrierConfig::isArrayOfLangs( array_keys( $v ) )
					&& CarrierConfig::isArrayOfStrings( $v )
					? CarrierConfig::sortArray( $v ) : wfMessage( 'zeroconfig-banner' );
			} );

		// Partner URL, do not link by default
		$this->check( 'bannerUrl', '',
			function( $v ) {
				return
					$v === '' || false !== filter_var( $v, FILTER_VALIDATE_URL )
						? null : wfMessage( 'zeroconfig-banner_url' );
			} );

		//'showLangs' => null,        // List of language codes to show on Zero page
		$this->check( 'showLangs', null,
			function( $v ) {
				if ( is_string( $v ) ) {
					$v = array( $v );
				}
				if ( CarrierConfig::isArray( $v, false )
					&& CarrierConfig::isArrayOfLangs( $v )
					&& count( $v ) > 0
				) {
					// Remove duplicates while preserving original order
					$v = array_unique( $v );
					ksort( $v );
					return $v;
				}
				return wfMessage( 'zeroconfig-show_langs' );
			} );

		//'whitelistedLangs' => null, // List of language codes to show banner on, or empty list to allow on all languages
		$showLangs = $this->config['showLangs'];
		$this->check( 'whitelistedLangs', null,
			function( $v ) use ( $showLangs ) {
				if ( is_string( $v ) ) {
					$v = array( $v );
				}
				if ( CarrierConfig::isArray( $v, false )
					&& CarrierConfig::isArrayOfLangs( $v )
					&& ( count( $v ) === 0 || count( array_diff( $showLangs, $v ) ) === 0 )
				) {
					if ( count( $v ) === 0 ) {
						// Empty list is the same as whitelist all languages
						return $v;
					}
					// Make $v in the same order as $showLangs, followed by alphabetical leftovers
					$v = array_unique( $v );
					$leftovers = array_diff( $v, $showLangs );
					sort( $leftovers );
					return array_merge( $showLangs, $leftovers );
				}
				return wfMessage( 'zeroconfig-whitelisted_langs' );
			} );

		// Orange Congo wanted to be able to override the 'kg' language name to 'Kikongo'
		$this->check( 'langNameOverrides', array(),
			function( $v ) {
				return CarrierConfig::isArray( $v, true )
					&& CarrierConfig::isArrayOfLangs( array_keys( $v ) )
					&& CarrierConfig::isArrayOfStrings( $v )
					? CarrierConfig::sortArray( $v ) : wfMessage( 'zeroconfig-lang_name_overrides' );
			} );

		// List of proxies supported by the carrier, defaults to none (empty list)
		$this->check( 'proxies', array(),
			function( $v ) use ( $isSaving ) {
				if ( CarrierConfig::isArray( $v, false ) ) {
					if ( $isSaving ) {
						// Remove duplicates while preserving original order
						$v = array_unique( $v );
						ksort( $v );
					}
					return $v;
				}
				return wfMessage( 'zeroconfig-proxies' );
			} );


		// Background banner color
		$this->check( 'background', '#E31230', $validateStr );

		// Foreground banner color
		$this->check( 'foreground', '#551011', $validateStr );

		// Banner font size override
		$this->check( 'fontSize', '', $validateStr );

		// Show "non-zero navigation" warning when clicking the banner
		$this->check( 'bannerWarning', true, $validateBool );

		// Zero rate images.
		// @BUG? does this have the same meaning as legacy "IMAGES_ON"?
		$this->check( 'showImages',  true, $validateBool );

		// Show the special zero page
		// default = ( count( 'showLangs' ) > 1 )
		$this->check( 'showZeroPage', true, $validateBool );

		// If carrier supports zero-rating HTTPS traffic
		$this->check( 'enableHttps', false, $validateBool );

		// List of IP CIDR blocks for this provider
		if ( $this->isSaving ) {
			$this->check( 'ips', '',
				function( $v ) {
					if ( is_string( $v ) ) {
						$v = array( $v );
					} elseif ( !CarrierConfig::isArray( $v, false ) || !CarrierConfig::isArrayOfStrings( $v ) ) {
						return wfMessage( 'zeroconfig-ips' );
					}

					// At some point we might remove FILTER_FLAG_NO_PRIV_RANGE
					// to allow local network testing
					$ipFlags = FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE;
					$optIPv4 = array( 'options' => array( 'min_range' => 1, 'max_range' => 32 ) );
					$optIPv6 = array( 'options' => array( 'min_range' => 1, 'max_range' => 128 ) );

					foreach ( $v as $cidr ) {
						$parts = explode( '/', $cidr, 3 );
						if ( count( $parts ) > 2
							|| false === filter_var( $parts[0], FILTER_VALIDATE_IP, $ipFlags )
						) {
							return wfMessage( 'zeroconfig-ips' );
						}
						if ( count( $parts ) === 2 ) {
							// CIDR block, second portion must be an integer within range
							// If the first part has ':', treat it as IPv6
							// Make sure there are no spaces in the block size
							$blockSize = filter_var(
								$parts[1], FILTER_VALIDATE_INT,
								strpos( $parts[0], ':' ) !== false ? $optIPv6 : $optIPv4 );
							if ( false === $blockSize || $parts[1] !== strval( $blockSize ) ) {
								return wfMessage( 'zeroconfig-ips' );
							}
						}
					}
					// Sort in natural order, but force sequential keys to prevent json to treat it as dictionary.
					natsort( $v );
					return array_values( $v );
				} );
		}

		// In case 'showZeroPage' is not set, the default depends on how many languages are shown.
		if ( isset( $this->defaultUsed['showZeroPage'] ) && !isset( $this->defaultUsed['showLangs'] ) ) {
			$this->config['showZeroPage'] = count( $this->config['showLangs'] ) > 1;
		}

		$errCount = count( $this->errors );
		if ( $errCount > 0 ) {
			$this->errorMessage = wfMessage( 'zeroconfig-errors_found' )->numParams( $errCount );
		}

		$this->config = array_merge( $this->config, $this->json );
		$this->unknownFields = $this->json;
		$this->json = null;
	}

	private function check( $field, $default, $validator ) {
		$res = $this->getValue( $field );
		if ( $res === null ) {
			$res = $default;
			$this->defaultUsed[$field] = true;
		} elseif ( is_object( $res ) && get_class( $res ) === 'Message' ) {
			$this->errors[$field] = $res;
			return;
		}
		$err = call_user_func( $validator, $res );
		if ( $err !== null ) {
			if ( is_object( $err ) && get_class( $err ) === 'Message' ) {
				$this->errors[$field] = $err;
				return;
			} else {
				$res = $err;
			}
		}
		$this->config[$field] = $res;
		if ( $res === $default ) {
			$this->sameAsDefault[$field] = true;
		}
	}

	/**
	 * Get field value from the json object
	 * This function may return an error message in case there are keys that only differ in their casing
	 * @param string $field field name
	 * @return Message|null
	 */
	private function getValue( $field ) {
		$value = null;
		$duplicates = array();

		// check for exact field name match
		if ( array_key_exists( $field, $this->json ) ) {
			$duplicates[] = $field;
			$value = $this->json[$field];
			unset( $this->json[$field] );
		}

		// check for other casing of the field name
		foreach ( $this->json as $k => $v ) {
			if ( 0 === strcasecmp( $k, $field ) ) {
				$duplicates[] = $k;
				$value = $this->json[$k];
				unset( $this->json[$k] );
			}
		}
		if ( count( $duplicates ) > 1 ) {
			return wfMessage( 'zeroconfig-duplicate_field' );
		}
		return $value;
	}

	static function isArray( $arr, $isAssoc ) {
		if ( !is_array( $arr ) ) {
			return false;
		}
		$strCount = count( array_filter( array_keys( $arr ), 'is_string') );
		return ( $isAssoc && $strCount === count( $arr ) ) ||
			( !$isAssoc && $strCount === 0 );
	}

	static function isArrayOfStrings( $arr ) {
		return count( $arr ) === count( array_filter( $arr, 'is_string' ) );
	}

	static function isArrayOfLangs( $arr ) {
		$filter = function( $v ) { return \Language::isValidCode( $v ); };
		return count( $arr ) === count( array_filter( $arr, $filter ) );
	}

	static function sortArray( $arr ) {
		// Sort so that the values are sorted alphabetically except 'en' which will go as first
		uksort( $arr,
			function( $a, $b ) {
				if ( $a === $b ) {
					return 0;
				} elseif ( $a === 'en' ) {
					return -1;
				} elseif ( $b === 'en' ) {
					return 1;
				} else {
					return strcasecmp($a, $b);
				}
			} );
		return $arr;
	}

	static function normalizeAdmins( $admins ) {
		$admins = array_unique( $admins );
		sort( $admins );
		return array_values( $admins );
	}
}
