<?php

namespace Extensions\ZeroRatedMobileAccess;
use FormatJson;
use Http;
use ObjectCache;
use stdClass;


/**
 * Represents a json blob on a remote wiki.
 * Handles retrieval (via HTTP) and local caching.
 * This code was adapted from extensions/EventLogging/includes/RemoteSchema.php
 * @note When we switch to PHP 5.4, add 'implements JsonSerializable'
 */
class CarrierConfigStore {

	const LOCK_TIMEOUT = 20;

	var $carrierId;
	var $cache;
	var $http;
	var $key;
	var $content = false;
	var $isCached = false;

	/**
	 * Constructor.
	 * @param string $carrierId Telecom ID (X-CS Header)
	 * @param ObjectCache $cache: (optional) cache client.
	 * @param Http $http: (optional) HTTP client.
	 */
	function __construct( $carrierId, $cache = NULL, $http = NULL ) {
		$this->carrierId = $carrierId;
		$this->cache = $cache ?: wfGetCache( CACHE_ANYTHING );
		$this->http = $http ?: new Http();
		$this->key = 'ZeroRatedMobileAccess:' . $carrierId;
	}


	/**
	 * Retrieves content.
	 * @return array|bool: Content array or false if irretrievable.
	 */
	public function get() {
		if ( !$this->isCached ) {
			// Get it from memcached
			$value = $this->memcGet();
			if ( $value === false ) {
				// Get it from HTTP
				$this->content = $this->httpGet();
				// Store it in memcached
				// Empty string is the magic value that means invalid config
				$value = $this->content === false ? '' : $this->content;
				$this->memcSet( $value );
			} elseif ( $value === '' ) {
				// Invalid ID was cached
				$this->content = false;
			} else {
				// Content was cached
				$this->content = $value;
			}
			$this->isCached = true;
		}
		return $this->content;
	}


	/**
	 * Retrieves content from memcached.
	 * @return array|bool Carrier config or false if not in cache.
	 */
	private function memcGet() {
		global $wgZeroRatedMobileAccessDisableCache;
		if ( $wgZeroRatedMobileAccessDisableCache ) {
			return false;
		}
		return $this->cache->get( $this->key );
	}


	/**
	 * Store content in memcached.
	 */
	private function memcSet( $value ) {
		global $wgZeroRatedMobileAccessDisableCache;
		if ( $wgZeroRatedMobileAccessDisableCache ) {
			return true;
		}
		return $this->cache->set( $this->key, $value );
	}


	/**
	 * Delete any cached information related to this config
	 */
	public function resetCache() {
		global $wgZeroRatedMobileAccessDisableCache;
		if ( $wgZeroRatedMobileAccessDisableCache ) {
			return;
		}
		$this->cache->delete( $this->key );
		$this->cache->delete( $this->key . ':lock' );
	}


	/**
	 * Acquire a mutex lock for HTTP retrieval.
	 * @return bool: Whether lock was successfully acquired.
	 */
	private function lock() {
		global $wgZeroRatedMobileAccessDisableCache;
		if ( $wgZeroRatedMobileAccessDisableCache ) {
			return true;
		}
		return $this->cache->add( $this->key . ':lock', 1, self::LOCK_TIMEOUT );
	}

	/**
	 * Retrieves the config using HTTP.
	 * Uses a memcached lock to avoid cache stampedes.
	 * @return array|boolean: config or false if unable to fetch.
	 */
	private function httpGet() {
		wfProfileIn( __METHOD__ );
		$pageInfo = false;
		if ( $this->lock() ) {
			$articleName = 'Zero:' . $this->carrierId;
			$query = array(
				'action' => 'query',
				'prop' => 'info|flagged',
				'format' => 'json',
				'titles' => $articleName,
			);
			$pageInfo = $this->getPageFromApi( $query );
			if ( $pageInfo !== false ) {
				$revId = array_key_exists ( 'flagged', $pageInfo ) ?
					$pageInfo['flagged']['stable_revid'] :
					$pageInfo['lastrevid'];

				$query = array(
					'action' => 'query',
					'prop' => 'revisions',
					'rvprop' => 'content',
					'format' => 'json',
					'revids' => $revId,
				);
				$pageInfo = $this->getPageFromApi( $query );
			}
		}
		wfProfileOut( __METHOD__ );
		return $pageInfo === false ? false : $pageInfo['revisions'][0]['*'];
	}

	/**
	 * Returns an object containing serializable properties.
	 * @implements JsonSerializable
	 */
	public function jsonSerialize() {
		return array(
			'Zero'  => $this->get() ?: new stdClass(),
		);
	}

	/** Given a legal set of API parameters, return page from API
	 * @param array $query
	 * @return bool|mixed
	 */
	private function getPageFromApi( $query )
	{
		global $wgZeroRatedMobileAccessApiUri;
		if ( $wgZeroRatedMobileAccessApiUri === false ) {
			wfLogWarning( '$wgZeroRatedMobileAccessApiUri is not set' );
			return false;
		}
		$apiUri = wfAppendQuery( $wgZeroRatedMobileAccessApiUri, $query );
		$revInfo = json_decode( $this->http->get( $apiUri, self::LOCK_TIMEOUT * 0.8 ), true );

		if ( !isset( $revInfo['query']['pages'] ) ) {
			return false;
		}
		$pages = $revInfo['query']['pages'];
		if ( !is_array( $pages ) || count( $pages ) !== 1 ) {
			return false;
		}
		$pageInfo = reset( $pages ); // get the only element of the array
		if ( array_key_exists( 'missing', $pageInfo ) ) {
			return false;
		}
		return $pageInfo;
	}
}
