<?php

namespace Extensions\ZeroRatedMobileAccess;


/**
 * Configuration page hook handlers for ZeroRatedMobileAccess extension.
 *
 * @file
 * @ingroup Extensions
 */
class ConfigPageHooks {

	/**
	 * Initialize state
	 * @param $output
	 * @param $page
	 * @param $title \Title
	 * @param $user
	 * @param $request
	 * @param $helper
	 * @return bool
	 */
	public static function onMediaWikiPerformAction( $output, $page, $title, $user, $request, $helper ) {
		if ( $title->inNamespace( NS_ZERO ) ) {
			PageRenderingHooks::getState( $output );
		}
		return true;
	}

	/**
	 * Declares JSON as the code editor language for Zero: pages.
	 * This hook only runs if the CodeEditor extension is enabled.
	 * @param \Title $title
	 * @param string &$lang Page language.
	 * @return bool
	 */
	static function onCodeEditorGetPageLanguage( $title, &$lang ) {
		if ( $title->inNamespace( NS_ZERO ) ) {
			$lang = 'json';
		}
		return true;
	}

	/**
	 * Validates that the revised contents are valid JSON.
	 * If not valid, rejects edit with error message.
	 * @param \IContextSource $context
	 * @param ZeroConfigContent $content
	 * @param \Status $status
	 * @param string $summary Edit summary provided for edit.
	 * @param \User $user
	 * @param bool $minoredit
	 * @return bool
	 */
	static function onEditFilterMergedContent( $context, $content, $status, $summary, $user, $minoredit ) {
		wfProfileIn( __METHOD__ );
		$title = $context->getTitle();
		if ( $title && $title->inNamespace( NS_ZERO ) ) {
			$content->getCarrierConfig()->setErrorMessageStatus( $status );
		}
		wfProfileOut( __METHOD__ );
		return true;
	}

	/**
	 * Adds CSS for pretty-printing configuration on NS_ZERO pages.
	 * @param \OutputPage &$out
	 * @param \Skin &$skin
	 * @return bool
	 */
	static function onBeforePageDisplay( &$out,
										 /** @noinspection PhpUnusedParameterInspection */ &$skin ) {
		$title = $out->getTitle();
		if ( $title->inNamespace( NS_ZERO ) ) {
			$out->addModules( 'zero.config' );
		}
		return true;
	}

	/**
	 * Reset cache
	 * @see https://www.mediawiki.org/wiki/Manual:Hooks/PageContentSaveComplete
	 */
	public static function onPageContentSaveComplete( $article, $user, $content, $summary,
													  $isMinor, $isWatch, $section, $flags, $revision, $status, $baseRevId ) {

		/** @var \WikiPage $article */
		$title = $article->getTitle();
		if ( $title && $title->inNamespace( NS_ZERO ) ) {
			$store = new CarrierConfigStore( $title->getText() );
			$store->resetCache();
		}
		return true;
	}

	/**
	 * Prohibit editing of Zero namespace unless the user is either in zero-edit group
	 * or the user is listed as one of the admins
	 * @param \Title $title
	 * @param \User $user
	 * @param string $action
	 * @param $result
	 * @return bool
	 */
	public static function onUserCan( &$title, &$user, $action, &$result = null ) {
		// We only care about non-reading rights for NS_ZERO namespace
		if ( $title->inNamespace( NS_ZERO ) && $action !== 'read' ) {
			if ( $user->isAllowed( 'zero-edit' ) ) {
				return true; // Allow global admins (users with 'zero-edit' right)
			} elseif ( $action === 'edit' ) {
				// onUserCan is called multiple times - cache the list of admins
				static $sTitle = null, $sAdmins;
				if ( $title !== $sTitle ) {
					$sTitle = $title;
					$sAdmins = false;
					// Partner admins can only edit existing page
					$wp = new \WikiPage( $title );
					/** @var ZeroConfigContent $content */
					$content = $wp->getContent( \Revision::RAW );
					if ( $content instanceof ZeroConfigContent ) {
						$cc = $content->getCarrierConfig();
						if ( !$cc->isError() ) {
							$config = $cc->getData();
							$sAdmins = $config['admins'];
						}
					}
				}
				if ( $sAdmins && in_array( $user->getName(), $sAdmins ) ) {
					return true; // User has been explicitly allowed for this page
				}
			}
			// Otherwise prohibit any zero namespace editing
			$result = false;
			return false;
		} else {
			return true;
		}
	}
}
