<?php

namespace Extensions\ZeroRatedMobileAccess;
use BaseTemplate;
use DOMElement;
use Html;
use IContextSource;
use Language;
use MinervaTemplate;
use MobileContext;
use OutputPage;
use RawMessage;
use ResourceLoader;
use SkinTemplate;
use SpecialPage;
use Title;
use WebRequest;

/**
 * Main class for ZeroRatedMobileAccess extension.
 *
 * This class only contains static members.
 * Static methods are used by the hooks and by the Special:ZeroRatedMobileAccess.
 * The state is based on the X-CS and other headers, and has an instance
 * of config object and and a few header-based cached values.
 *
 * @file
 * @ingroup Extensions
 */
class PageRenderingHooks {

	/**
	 * @param IContextSource $context
	 * @return PageRendering
	 */
	public static function getState( IContextSource $context = null ) {
		static $state = null;
		if ( $state === null ) {
			$state = new PageRendering( $context );
		}
		return $state;
	}

	/**
	 * @param $out
	 * @return bool
	 */
	public static function onBeforePageDisplay( &$out ) {
		return self::getState( $out )->onBeforePageDisplay( $out );
	}

	/**
	 * @param MobileContext $context
	 * @param \MobileFormatter $formatter
	 * @return bool
	 */
	public static function onMobileFrontendBeforeDOM( MobileContext $context, \MobileFormatter $formatter ) {
		return self::getState( $context )->onMobileFrontendBeforeDOM( $formatter );
	}

	/**
	 * @param MinervaTemplate $template
	 * @return bool
	 */
	public static function onMinervaPreRender( MinervaTemplate $template ) {
		return self::getState()->onMinervaPreRender( $template );
	}

	/**
	 * @param SkinTemplate $sk
	 * @param $notice
	 * @return bool
	 */
	public static function onGetMobileNotice( SkinTemplate $sk, &$notice ) {
		return self::getState( $sk )->onGetMobileNotice( $notice );
	}

	/**
	 * @param $subdomainTokenReplacement
	 * @return bool
	 */
	public static function onGetMobileUrl( &$subdomainTokenReplacement ) {
		return self::getState()->onGetMobileUrl( $subdomainTokenReplacement );
	}
}

/**
 * Class PageRendering handles all the hook callbacks in a stateful maner
 * @package Extensions\ZeroRatedMobileAccess
 */
class PageRendering {

	private static $zerodotParent = '.zero.wikipedia.org/';
	private static $mdotParent = '.m.wikipedia.org/';

	/** @var \WebRequest */
	private $request;
	/** @var \Title */
	private $title;
	/** @var \SplBool */
	private $isZeroSpecial;
	/** @var \SplBool */
	private $isZeroSubdomain;
	/** @var \SplBool */
	private $isWap = false;
	/** @var \Language */
	private $lang;

	/**
	 * @param IContextSource $context
	 */
	public function __construct( IContextSource $context = null ) {
		if ( $context === null ) {
			// wfDebug( __METHOD__ . "Constructor with null \$context was called, using globals for sanity\n" );
			// Current common case: GetMobileUrl hook - needs to be updated to get context
			global $wgRequest, $wgOut;
			$this->request = $wgRequest;
			$this->title = $wgOut->getTitle();
			$this->lang = $wgOut->getContext()->getLanguage();
		} else {
			$this->request = $context->getRequest();
			$this->title = $context->getTitle();
			$this->lang = $context->getLanguage();
		}
		$this->isZeroSpecial = $this->title ? $this->title->isSpecial( 'ZeroRatedMobileAccess' ) : false;

		// Allow URL override for testing purposes
		$sd = $this->request->getVal( 'X-SUBDOMAIN' );
		if ( $sd === null ) {
			$sd = $this->request->getHeader( 'X-SUBDOMAIN' );
		}
		if ( $sd ) {
			$this->isZeroSubdomain = strcasecmp( $sd, 'ZERO' ) === 0;
		} else {
			// Either running from root redirector or from shell
			$this->isZeroSubdomain = false;
			$urlParts = wfParseUrl( $this->request->detectServer() );
			if ( isset( $urlParts['host'] ) ) {
				$urlParts = explode( '.', $urlParts['host'], 3 );
				if ( count( $urlParts ) >= 2 ) {
					$this->isZeroSubdomain = strcasecmp( $urlParts[0], 'ZERO' ) === 0 ||
											 strcasecmp( $urlParts[1], 'ZERO' ) === 0;
				};
			}
		}
	}

	// TODO: Clean up HTML concatenation. Review for any non-escaped user input.

	/**
	 * GetMobileUrl hook handler - replaces *.m.* with *.zero.* if the current site is zero.
	 *
	 * @param $subdomainTokenReplacement string
	 * @return bool
	 */
	public function onGetMobileUrl( &$subdomainTokenReplacement ) {
		if ( $this->isZeroSubdomain ) {
			$subdomainTokenReplacement = 'zero.';
		}
		return true;
	}

	/**
	 * Rewrite DOM to replace content's external and image links with links to the special redirect page.
	 * @param \MobileFormatter $formatter
	 * @return bool
	 */
	public function onMobileFrontendBeforeDOM( \MobileFormatter $formatter ) {

		if ( !$this->isZeroSite() ) {
			return true;
		}
		$config = $this->getConfig();
		if ( $config === null ) {
			return true;
		}
		if ( $this->isZeroSpecial ) {
			return true;
		}

		wfProfileIn( __METHOD__ );
		$isFilePage = $this->title->inNamespace( NS_FILE );
		$isZerodotNonFilePage = !$isFilePage && $this->isZeroSubdomain;
		$doc = $formatter->getDoc();

		// ZERO site - replace all images with links to images
		if ( $isZerodotNonFilePage ) {
			$tagToReplaceNodes = $doc->getElementsByTagName( 'img' );
			$tagToReplaceNodesCollection = array();
			/* @var $tagToReplaceNode DOMElement */
			foreach ( $tagToReplaceNodes as $tagToReplaceNode ) {
				if ( $tagToReplaceNode ) {
					$alt = $tagToReplaceNode->getAttribute( 'alt' );
					$spanNodeText = wfMessage(
						'zero-rated-mobile-access-click-to-view-image',
						lcfirst( substr( $alt, 0, 40 ) )
					)->text();
					$spanNode = $doc->createElement( "span", str_replace( "&", "&amp;", $spanNodeText ) );
					if ( $alt ) {
						$spanNode->setAttribute( 'title', $alt );
					}
					$tagToReplaceNodesCollection[] = array( 'tagToReplaceNode' => $tagToReplaceNode, 'spanNode' => $spanNode );
				}
			}
			foreach ( $tagToReplaceNodesCollection as $element ) {
				$element['tagToReplaceNode']->parentNode->replaceChild( $element['spanNode'], $element['tagToReplaceNode'] );
			}
		}

		if ( $isZerodotNonFilePage ) {
			// MW normalizes all links to the main namespace, ignoring all alternative ones
			$ns = $this->lang->convertNamespace( NS_FILE );
			$filePrefix = '#^/wiki/' . $ns . ':.#';
		}
		/* @var $link DOMElement */
		foreach ( $doc->getElementsByTagName( 'a' ) as $link ) {
			$href = $link->getAttribute( 'href' );
			if ( !$href ) {
				continue; // Missing HREF
			}
			$ch0 = substr( $href, 0, 1 );
			if ( $ch0 === '#' ) {
				continue; // Anchor is never checked
			}
			$ch1 = substr( $href, 1, 1 );
			$class = $link->getAttribute( 'class' );
			// Any link to image (class=image), or any non-local link, or, in a zerodot
			// article, any direct File: link will be redirected.
			// Local links start with either '?...' or '/...', but not '//...'
			// Local links to File: are indicative of a media type.
			if ( $class !== 'image' &&
				( $ch0 === '?' || ( $ch0 === '/' && $ch1 !== '/' ) ) &&
				!( $isZerodotNonFilePage && preg_match( $filePrefix, $href ) )
			) {
				continue;
			}
			$link->setAttribute( 'href', $this->makeRedirect( $href ) );
		}

		wfProfileOut( __METHOD__ );
		return true;
	}

	/**
	 * Invoked by MinervaTemplate to add banners
	 * @param MinervaTemplate $template
	 * @return bool
	 */
	public function onMinervaPreRender( MinervaTemplate $template ) {
		// If the page is Special:ZeroRatedMobileAccess just wipe out the menu "hamburger".
		if ( $this->isZeroSpecial ) {
			$template->set( 'menuButton', '' );
		}

		// TODO: alias $template->data where appropriate
		$bannersSupported = array_key_exists( 'banners', $template->data );

		$banner = $this->createBanner();
		if ( $banner ) {
			if ( $bannersSupported ) {
				$template->set( 'banners', array( $banner ) );
			}
			if ( $this->getConfig() !== null ) {
				if ( $bannersSupported ) {
					$this->rewriteLangLinks( $template );
				}
				// Add warnings to footer links
				$this->addWarning( $template, 'mobile-switcher' );
				$this->addWarning( $template, 'mobile-license' );
				$this->addWarning( $template, 'privacy' );
				$this->addWarning( $template, 'terms-use' );
			}
		}
		return true;
	}

	/**
	 * Invoked by SkinMobileWML only to add banners
	 * @param $notice
	 * @return bool
	 */
	public function onGetMobileNotice( &$notice ) {
		$this->isWap = true;
		$banner = $this->createBanner( true );
		if ( $banner ) {
			$notice .= $banner;
		}
		return true;
	}

	/**
	 * BeforePageDisplayMobile hook handler
	 *
	 * @param $out OutputPage
	 * @return bool
	 */
	public function onBeforePageDisplay( &$out ) {
		if ( !$this->isZeroSite() ) {
			return true;
		}

		wfProfileIn( __METHOD__ );

		// @FIXME: Debug checks - should be remove once all issues are gone
		$hasObsoleteFlags =
			$this->request->getCheck( 'acceptbilling' ) ||
			$this->request->getCheck( 'renderwarning' ) ||
			$this->request->getCheck( 'renderZeroRatedBanner' ) ||
			$this->request->getCheck( 'renderZeroRatedRedirect' );
		if ( $hasObsoleteFlags || $this->getConfigId() === null ) {
			$warn = $hasObsoleteFlags ? 'zero params in query' : '';
			if ( $this->isZeroSubdomain ) {
				$referer = $this->request->getHeader( 'REFERER' );
				if ( strpos( $referer, 'zero.wikipedia.org' ) !== false ) {
					$warn .= ' && zero referer';
				}
			}
			$ua = $this->request->getHeader( 'USER-AGENT' );
			if ( $warn !== ''
				&& $this->request->getHeader( 'FROM' ) !== 'googlebot(at)googlebot.com'
				&& strpos( $ua, 'Googlebot' ) === false
				&& strpos( $ua, 'bingbot' ) === false
				&& strpos( $ua, 'YandexBot' ) === false
				&& strpos( $ua, 'msnbot' ) === false
			) {
				self::logDebug( '!config' . $warn );
			}
		}

		if ( $this->getConfig() !== null || $this->isZeroSubdomain ) {
			$out->addModuleStyles( 'mobile.zero.styles' );
			$out->addModules( 'mobile.zero' );
		}

		$out->addVaryHeader( 'X-CS' );
		$out->addVaryHeader( 'X-Subdomain' );
		$out->addVaryHeader( 'X-Forwarded-By' );

		// If zerodot isn't supported here and the user isn't already on
		// Special:ZeroRatedMobileAccess, send the user to Special:ZeroRatedMobileAccess
		// and include the original URL so that the user may continue to the m. server.
		if ( !$this->isZeroSpecial && $this->isZeroSubdomain && $this->getConfig() === null ) {
			$info = $this->getWikiInfo();
			$url = $this->getStartPageUrl( $info[0], false, true );
			$url = wfAppendQuery( $url, array( 'from' => $this->request->getFullRequestURL() ) );
			$out->redirect( $url );
			$out->output();
		}

		wfProfileOut( __METHOD__ );
		return true;
	}

	/**
	 * @param bool $wap
	 * @return string with banner html
	 */
	private function createBanner( $wap = false ) {
		global $wgZeroRatedMobileAccessEnableESI;
		// fixme: temporary until we stabilize it, then remove X-FORCE-ESI check
		if ( !$this->isZeroSpecial &&
			( $wgZeroRatedMobileAccessEnableESI === true || $this->request->getHeader( 'X-FORCE-ESI' ) )
		) {
			# Add an 'Enable-ESI' header for varnish
			$this->request->response()->header( 'Enable-ESI: 1' );

			$query = 'renderesibanner=1';
			if ( $this->title->inNamespace( NS_FILE ) ) {
				$query .= '&img=1';
			}
			if ( $wap ) {
				$query .= '&wap=1';
			}
			$url = self::getSpecial()->getLocalURL( $query );
			$esi = '<esi:include' . Html::expandAttributes( array( 'src' => $url, 'onerror' => 'continue' ) ) . '/>';

			return $esi;
		}
		$config = $this->getConfig();
		if ( $config === null ) {
			// @fixme logic: In case of no X-CS, should we ignore all redirect attempts?
			return '';
		}

		if ( $this->isZeroSpecial ) {
			$redir = $this->getRedirectInfo();
			if ( array_key_exists( 'warn', $redir ) || array_key_exists( 'redirect', $redir ) ) {
				return '';
			}
		}
		return $this->renderBanner(
			$config, null, null, $wap, $this->title->inNamespace( NS_FILE ), $this->isZeroSpecial );
	}

	/**
	 * @param $config
	 * @param string|null $langCode If null, use site's default.
	 * @param string $sitename
	 * @param bool $wap
	 * @param bool $isImage
	 * @param bool $undismissable
	 * @return string
	 */
	public function renderBanner(
		$config, $langCode, $sitename = null, $wap = false, $isImage = false, $undismissable = false
	) {
		if ( $isImage && !$config['showImages'] ) {
			return '';
		}
		$banner = self::getBannerText( $config, $langCode, $sitename );
		if ( !$banner ) {
			return '';
		}
		if ( !$wap ) {
			$background = $config['background'];
			$foreground = $config['foreground'];
		}
		$billingURL = $config['bannerUrl'];
		if ( $billingURL ) {
			if ( $config['bannerWarning'] ) {
				$billingURL = $this->makeRedirect( $billingURL );
			}
			$attr = array( 'href' => $billingURL );
			if ( !$wap ) {
				$attr['style'] = "color:$foreground;";
			}
			$banner = Html::rawElement( 'a', $attr, $banner );
		}

		if ( $wap ) {
			$banner = $banner !== '' ? Html::rawElement( 'p', null, $banner ) : '';
		} else {
			$attr = array(
				'class' => 'mw-mf-message',
				'id' => 'zero-rated-banner-text',
			);
			$style = $config['fontSize'] !== '' ? 'font-size:' . $config['fontSize'] . ';' : '';
			if ( !$billingURL ) {
				$style .= "color:$foreground;";
			}
			if ( $style !== '' ) {
				$attr['style'] = $style;
			}
			$banner = Html::rawElement( 'span', $attr, $banner );
			if ( !$undismissable ) {
				$dismissTitle = wfMessage( 'zero-rated-mobile-access-dismiss-notification' )->escaped();
				$dismiss = Html::rawElement(
					'button',
					array( 'class' => 'notify-close',
						'style' => "background:$background;",
						'title' => $dismissTitle ),
					Html::rawElement(
						'span',
						array(
							'class' => 'notify-close-x',
							'style' => "background:$background;border-color:$foreground;color:$foreground;" ),
						'&#120;'
					)
				);
			} else {
				$dismiss = '';
			}
			$banner = Html::rawElement(
				'div',
				array(
					'id' => 'zero-rated-banner',
					'class' => 'mw-mf-banner',
					'style' => "background:$background;color:$foreground;" ),
				$dismiss . $banner );
		}

		return $banner;
	}

	/**
	 * If a particular language could cause a charge, send user to an interstitial.
	 *
	 * @param BaseTemplate $template
	 * @return bool
	 */
	private function rewriteLangLinks( BaseTemplate $template ) {

		if ( isset( $template->data['language_urls'] ) ) {
			$languageUrls = $template->data['language_urls'];
			if ( is_array( $languageUrls ) && count( $languageUrls ) > 0 ) {
				foreach ( $languageUrls as &$lang ) {
					$link = $lang['href'];
					if ( $this->isZeroSubdomain ) {
						$link = str_replace( self::$zerodotParent, self::$mdotParent, $link );
					}
					$lang['href'] = $this->makeRedirect( $link );
				}
				$template->set( 'language_urls', $languageUrls );
			}
		}
		return true;
	}

	/**
	 * @param $config
	 * @param string $langCode
	 * @param string $sitename
	 * @return bool|String
	 */
	public static function getBannerText( $config, $langCode, $sitename ) {
		wfProfileIn( __METHOD__ );

		if ( $langCode === null ) {
			global $wgLang;
			$lang = $wgLang;
		} else {
			$lang = Language::factory( $langCode );
		}
		$name = self::pickLocalizedString( $config['name'], $lang );
		if ( $name === false ) {
			wfProfileOut( __METHOD__ );
			return false;
		}
		$nameMsg = new RawMessage( $name );

		if ( array_key_exists( $lang->getCode(), $config['banner'] ) ) {
			$linkText = $config['banner'][$lang->getCode()];
		} else {
			$linkText = wfMessage( 'zero-rated-mobile-access-banner-text' )->inLanguage( $lang )->plain();
		}
		if ( $sitename !== null ) {
			// Rendering for the zero-config page, fake {{SITENAME}}
			$linkText = str_replace( '{{SITENAME}}', $sitename, $linkText );
		}
		$linkTextMsg = new RawMessage( $linkText );
		$res = $linkTextMsg->inLanguage( $lang )->rawParams( $nameMsg->inLanguage( $lang )->escaped() )->escaped();

		wfProfileOut( __METHOD__ );
		return $res;
	}

	/**
	 * Return Carrier ID (X-CS) value or null if no value was set
	 * @return null|string
	 */
	public function getConfigId() {
		if ( $this->configId === false ) {
			// Allow URL override of the X-CS parameter for testing purposes
			$id = $this->request->getVal( 'X-CS' );
			if ( $id === null ) {
				$id = $this->request->getHeader( 'X-CS' );
			}
			if ( $id === '(null)' || !$id ) {
				$id = null;
			}
			$this->configId = $id;
		}
		return $this->configId;
	}
	private $configId = false;

	/**
	 * Returns cached carrier configuration based on X-CS header or query parameter
	 * @param bool $allowDisabled if true, a valid but disabled config object would still be returned
	 * @return array|null Carrier configuration or null if it doesn't exist or is disabled
	 */
	public function getConfig( $allowDisabled = false ) {

		// Cached configuration array or null if missing or disabled.
		if ( $this->config === false ) {
			wfProfileIn( __METHOD__ );
			$config = null;
			$configDisabled = null;
			// Allow URL override of the X-CS parameter for testing purposes
			$id = $this->getConfigId();
			if ( $id !== null ) {
				$store = new CarrierConfigStore( $id );
				$text = $store->get();
				if ( $text !== false ) {
					$conf = new CarrierConfig( $text );
					if ( !$conf->isError() ) {
						$config = $conf->getData();
						$configDisabled = $config;
						if ( !$this->isZeroEnabled( $config ) ) {
							$config = null;
						}
					}
				}
			}
			$this->config = $config;
			$this->configDisabled = $configDisabled;
			wfProfileOut( __METHOD__ );
		}
		return $allowDisabled ? $this->configDisabled : $this->config;
	}
	private $config = false;
	private $configDisabled = false;

	/**
	 * As part of getConfig(), determine if zero should be enabled
	 * for this site based on configuration
	 * @param array $config
	 * @return bool
	 */
	private function isZeroEnabled( $config ) {
		if ( !$config['enabled'] && !$this->request->getCheck( 'X-CS' )) {
			// If the configuration is disabled, pretend like it doesn't exist,
			// unless the configuration was specified with the X-CS request parameter
			return false;
		}

		$xfb = $this->request->getHeader( 'X-Forwarded-By' );
		if ( $xfb !== false && !in_array( $xfb, $config['proxies'] ) ) {
			return false;
		}

		// Check if the wiki is in the allowed site list and languages list
		list( $lang, $subdomain, $site ) = $this->getWikiInfo();
		if ( !in_array( $subdomain . '.' . $site, $config['sites'] ) ) {
			return false;
		}
		$freeLangs = $config['whitelistedLangs'];
		if ( 0 !== count( $freeLangs ) && !in_array( $lang, $freeLangs ) ) {
			return false;
		}
		if ( !$config['enableHttps'] && $this->request->getProtocol() === 'https' ) {
			return false;
		}
		return true;
	}

	/**
	 * @return array with three values - language code (en), subdomain (m), and site (wikipedia)
	 */
	public function getWikiInfo() {
		static $info = null;
		if ( $info === null ) {
			global $wgConf, $wgDBname;
			list( $site, $langCode ) = $wgConf->siteFromDB( $wgDBname );
			if ( $langCode === '' || $langCode === 'test' ) {
				$langCode = 'en'; // Useful for debugging, should never be the case in production
			}
			if ( !$site ) {
				$site = 'wikipedia';
			}
			$info = array( $langCode, $this->isZeroSubdomain ? 'zero' : 'm', $site );
		}
		return $info;
	}

	/**
	 * Returns true if X-SUBDOMAIN header is set to 'ZERO'
	 * @return bool
	 */
	public function isZeroSubdomain() {
		return $this->isZeroSubdomain;
	}

	/**
	 * Determine if this could be a Zero-related site shown with a mobile device
	 * Must be wikipedia, with mobile view enabled
	 * @return bool
	 */
	private function isZeroSite() {
		static $isZero = null;
		if ( $isZero === null ) {
			global $wgEnableZeroRatedMobileAccessTesting;
			$info = $this->getWikiInfo();
			// @FIXME: this check should only see if this is a mobileview, the rest is a check in partner's config
			$isZero = ( $info[2] === 'wikipedia' || $wgEnableZeroRatedMobileAccessTesting )
				&& MobileContext::singleton()->shouldDisplayMobileView();
		}
		return $isZero;
	}

	/**
	 * For the special zero page when used as a redirector,
	 * determine if redirection can be done automatically or a warning should be displayed.
	 * Returns empty array if no redirection,
	 * array( 'redirect' => $url, 'code' => '302' ) to auto-redirect
	 * array( 'softredirect' => $url ) in case there is no valid config (X-CS)
	 * array( 'warn' => 'external' or 'file', 'from' => $fromUrl, 'to' => $toUrl ) to show warning
	 */
	public function getRedirectInfo() {
		if ( $this->redirectInfo === null ) {
			$this->redirectInfo = $this->makeRedirectInfo();
		}
		return $this->redirectInfo;
	}
	private $redirectInfo = null;

	/**
	 * See getRedirectInfo() for description
	 * @return array
	 */
	private function makeRedirectInfo() {
		if ( $this->request->getCheck( 'isroot' ) ) {
			$url = $this->getLandingRedirect();
			return array( 'redirect' => $url, 'code' => '302' );
		}
		$config = $this->getConfig();
		$from = $this->request->getVal( 'from' );
		$toUrl = $this->request->getVal( 'to' );
		if ( $toUrl === null || $from === null ) {
			// This is not an external link or file redirect, see if we need to redirect anywhere
			$langCode = null;
			$forceMobile = false;
			$isLanding = false;
			if ( $config !== null ) {
				//
				if ( !$config['showZeroPage'] ) {
					$langCode = $config['showLangs'][0];
				}
			} elseif ( !$this->isZeroSubdomain ) {
				// Unidentified carrier, M subdomain redirect to main page
				// TODO: discuss if we ever want to send 'm.wikipedia.org/wiki/Special:ZRMA' -> 'zero.'
				// TODO: in case provider exists and they have only whitelisted zero.
				$info = $this->getWikiInfo();
				$langCode = $info[0];
			} else {
				$config = $this->getConfig( true );
				if ( $config !== null ) {
					// The traffic is from a carrier, but this specific domain is not zero-rated
					// Try to redirect to the free traffic if possible
					$info = $this->getWikiInfo();
					$sites = $config['sites'];
					if ( in_array( 'm.' . $info[2], $sites ) ) {
						$langCode = $info[0];
						$forceMobile = true;
					}
				}
				// else - do not redirect if this is "*.zero.wikipedia.org" - need to show red warning
			}

			if ( $langCode !== null ) {
				$url = $this->getStartPageUrl( $langCode, $forceMobile, $isLanding );
				return array( 'redirect' => $url, 'code' => '302' );
			}
			return array();
		}
		$fromTitle = Title::newFromText( $from );
		if ( !$fromTitle ) {
			self::logDebug( 'invalidParams' );
			return array();
		}
		if ( $config === null ) {
			self::logDebug( '!config' );
			return array( 'softredirect' => $toUrl );
		}
		$fromUrl = MobileContext::singleton()->getMobileUrl( $fromTitle->getFullURL() );
		$redir = false;
		$urlBits = wfParseUrl( $toUrl );
		$toHost = is_array( $urlBits ) && array_key_exists( 'host', $urlBits ) ? $urlBits['host'] : false;
		if ( $toHost ) {
			$toHost = strtolower( $toHost );
			// fixme: need to check $host against local wikipedia host name
			// fixme: this is a very rare case - a full URL to the local host
			// if ( strcasecmp( $toHost, getCurrentWikiHostName ) === 0 ) {
			// 	$toHost = false;
			// }
		}
		if ( $toHost ) {
			// Match (optional-language.)(subdomain.site).org
			if ( preg_match( '/^([^.]+\.)?([^.]+\.[^.]+)\.org$/', $toHost, $matches ) ) {
				// Another language in wikipedia
				$lang = $matches[1];
				$site = $matches[2];
				// see if the site is whitelisted, and if it is, make sure the language is.
				if ( in_array( $site, $config['sites'] ) ) {
					if ( $lang ) {
						$freeLangs = $config['whitelistedLangs'];
						$redir = count( $freeLangs ) == 0 || in_array( rtrim( $lang, '.' ), $freeLangs );
					} else {
						$redir = true; // there is no language, but the site is whitelisted, so don't warn
					}
				}
			}
			// else - external link, always warn
		} else {
			// there was no host - must be a local link to an image
			$redir = $config['showImages'];
		}
		if ( $redir ) {
			return array( 'redirect' => $toUrl, 'code' => '302' );
		}
		return array(
			'warn' => ( $toHost ? 'external' : 'file' ),
			'from' => $fromUrl, 'to' => $toUrl );
	}

	/**
	 * Format URL to the zero or m main page for the specific language (or current if null)
	 * @param string $langCode if given, target language, otherwise current one
	 * @param bool $forceMobile if true, will always use m. even on zero. sites
	 * @param bool $isLanding if true, will redirect to "Special:Zero" otherwise "Main Page"
	 * @return string
	 */
	public function getStartPageUrl( $langCode = null, $forceMobile = false, $isLanding = false ) {
		$info = $this->getWikiInfo();
		if ( $langCode === null ) {
			$langCode = $info[0];
		}
		$subdomain = $forceMobile ? 'm' : $info[1];
		$site = $info[2];
		$page = $isLanding ? self::getSpecial( false )->getPrefixedDBkey(): 'Main_Page';
		return sprintf( '//%s.%s.%s.org/wiki/%s', $langCode, $subdomain, $site, $page );
	}

	/**
	 * Find a message in a dictionary for the given language,
	 * or use language fallbacks if message is not defined.
	 * @param array $map Dictionary of languageCode => string
	 * @param Language $lang language object
	 * @return string|bool message from the dictionary or false if nothing found
	 */
	private static function pickLocalizedString( $map, $lang ) {
		$langCode = $lang->getCode();
		if ( array_key_exists( $langCode, $map ) ) {
			return $map[$langCode];
		}
		$fallbacks = $lang->getFallbackLanguages();
		if ( count( $fallbacks ) === 0 ) {
			$fallbacks = array( 'en' );
		}
		foreach ( $fallbacks as $l ) {
			if ( array_key_exists( $l, $map ) ) {
				return $map[$l];
			}
		}
		return false;
	}

	/**
	 * Create redirect URL to be inserted by DOM processing instead of all external links
	 * and other Zero-related redirects.
	 * @param string $url to redirect to in case the user accepts (or if its free for the user)
	 * @return String
	 */
	public function makeRedirect( $url ) {
		if ( $this->query === false ) {
			$this->query = 'from=' . $this->title->getPrefixedURL() . '&to=';
		}
		return self::getSpecial()->getLinkURL( $this->query . wfUrlencode( $url ) );
	}
	private $query = false;

	/**
	 * Adds parameters to URLs. Helper for onMinervaPreRender( BaseTemplate &$template )
	 * @param BaseTemplate $template
	 * @param string $templateValName The name of the template
	 */
	private function addWarning( BaseTemplate $template, $templateValName ) {
		$link = $template->get( $templateValName );
		$self = $this;
		$link = preg_replace_callback( '/href=([\'"])(.*?)\1/',
			function ( $matches ) use ( $self ) {
				$quoteChar = $matches[1];
				$href = $matches[2];
				$ch0 = substr( $href, 0, 1 );
				$ch1 = substr( $href, 1, 1 );
				// Local links start with either '#...', '?...' or '/...', but not '//...'
				if ( $ch0 !== '#' && $ch0 !== '?' && ( $ch0 !== '/' || $ch1 === '/' ) ) {
					return 'href=' .
						$quoteChar . $self->makeRedirect( htmlspecialchars_decode( $href ) ) . $quoteChar;
				} else {
					return $matches[0];
				}
			},
			$link
		);
		$template->set( $templateValName, $link );
	}

	/**
	 * Output debug info into Zero log
	 * @param string $dbg Debug message
	 */
	private static function logDebug( $dbg ) {
		static $printedHeaders = false;
		global $wgRequest;
		$dbg .= "\t" . $wgRequest->getIP() . "\t" . $wgRequest->getFullRequestURL();
		if ( !$printedHeaders ) {
			$headers = $wgRequest->getAllHeaders();
			ksort( $headers );
			$dbg .= "\nHeaders: " . print_r( $headers, true );
			$printedHeaders = true;
		}
		wfDebugLog( 'zero', $dbg );
	}

	/**
	 * Returned cached instance of the ZRMA special page
	 * @param bool $isLocalized - should the namespace "Special:" be localized
	 * @return Title Zero rated special page
	 */
	private static function getSpecial( $isLocalized = true ) {
		static $local = null, $global = null;
		if ( $isLocalized ) {
			if ( $local === null ) {
				$local = SpecialPage::getTitleFor( 'ZeroRatedMobileAccess' );
			}
			return $local;
		} else {
			if ( $global === null ) {
				$global = Title::makeTitle( NS_SPECIAL, 'ZeroRatedMobileAccess' );
			}
			return $global;
		}
	}

	/**
	 * Create redirect URL for m.* and zero.* root pages
	 * @return string
	 */
	public function getLandingRedirect() {
		$config = $this->getConfig();
		if ( $config === null ) {
			// Staying consistent with the current behaviour - redirect to en.*.*
			$langCode = 'en';
			// zero.* -> Landing, m.* -> Main Page
			$isLanding = $this->isZeroSubdomain;
		} else {
			$langCode = $config['showLangs'][0];
			$isLanding = $config['showZeroPage'];
		}
		$url = $this->getStartPageUrl( $langCode, false, $isLanding );
		return wfExpandUrl( $url );
	}
}
