<?php

namespace Extensions\ZeroRatedMobileAccess;

use FormatJson;
use Language;
use ParserOptions;
use Title;
use User;
use Xml;

/**
 * Represents the content of a JSON Zero Config article.
 * @file
 * @ingroup Extensions
 * @ingroup ZeroRatedMobileAccess
 *
 * @author Yuri Astrakhan <yurik@wikimedia.org>, based on Ori Livneh <ori@wikimedia.org> schema extension
 */
class ZeroConfigContent extends \TextContent {

	private $config;

	function __construct( $text ) {
		parent::__construct( $text, 'JsonZeroConfig' );
		$this->config = new CarrierConfig( $text, true ); // thorough validation
	}

	/**
	 * @return bool: Whether content is valid JSON Schema.
	 */
	function isValid() {
		return !$this->config->isError();
	}

	/**
	 * @return CarrierConfig
	 */
	function getCarrierConfig() {
		return $this->config;
	}

	/**
	 * Beautifies JSON prior to save.
	 * @param Title $title Title
	 * @param User $user User
	 * @param ParserOptions $popts
	 * @return ZeroConfigContent
	 */
	public function preSaveTransform( Title $title, User $user, ParserOptions $popts ) {
		if ( !$this->config->isValidJson() ) {
			// Invalid JSON - can't do anything with it
			return $this;
		}
		if ( !$this->config->isError() ) {
			// A sanity check just to make sure we don't inject non-authorized user. Might be unnecessary.
			if ( ConfigPageHooks::onUserCan( $title, $user, 'edit' ) && !$user->isAllowed( 'zero-edit' ) ) {
				// Make sure partner admins can't delete themselves from the admin list (prevents partner lockouts)
				$this->config->addAdmin( $user );
			}
			// There are no errors, normalize data (without defaults)
			$data = $this->config->getData( false );
		} else {
			// JSON is valid, but the data has errors
			$data = $this->config->getOriginalJson();
		}
		$formatted = FormatJson::encode( $data, true, FormatJson::ALL_OK );
		if ( $this->getNativeData() !== $formatted ) {
			return new ZeroConfigContent( $formatted );
		}
		return $this;
	}

	/**
	 * Generates HTML representation of content.
	 * @return string HTML representation.
	 */
	function getHtml() {
		$config = $this->config;
		if ( $config->isError() ) {
			return $config->getErrorMessage();
		} else {
			$data = $config->getData();
			return self::objectTable( $data ) . '<br />' . $this->renderBanners( $data );
		}
	}

	/**
	 * Constructs an HTML representation of a JSON object.
	 * @param Array $config
	 * @param int $level Tree level
	 * @param string $parentKey The name of the parent key
	 * @return string: HTML.
	 */
	private function objectTable( $config, $level = 0, $parentKey = null ) {
		wfProfileIn( __METHOD__ );

		$rows = array();
		$tdAttributes = array( 'class' => 'mw-zero-config-value' );

		foreach ( $config as $key => $val ) {
			if ( is_string( $key ) ) {
				$th = Xml::elementClean( 'th', null, $key );
			} else {
				$th = '';
			}
			if ( is_array( $val ) ) {
				if ( count( $val ) === 0 ) {
					if ( $level === 0 && $key === 'whitelistedLangs' ) {
						$td = Xml::element( 'td', $tdAttributes, wfMessage( 'zeroconfig-whitelisted_langs-all' ) );
					} else {
						$td = Xml::element( 'td', $tdAttributes, '' );
					}
				} elseif ( $level === 0 && ( $key === 'showLangs' || $key === 'whitelistedLangs' ) ) {
					$td = Xml::tags( 'td', $tdAttributes, implode( ', ', $val ) );
				} elseif ( $level === 0 && $key === 'admins' ) {
					$vals = array();
					foreach ( $val as $v ) {
						$url = Title::newFromText( $v, NS_USER )->getLinkURL();
						$vals[] = Xml::element( 'a', array( 'href' => $url ), $v );
					}
					$td = Xml::tags( 'td', $tdAttributes, implode( ', ', $vals ) );
				} else {
					$td = Xml::tags( 'td', null, $this->objectTable( $val, $level + 1, $key ) );
				}
			} else {
				if ( !is_string( $val ) ) {
					$val = FormatJson::encode( $val );
				}
				$td = Xml::elementClean( 'td', $tdAttributes, $val );
			}

			$attribs = null;
			switch ( $level ) {
				case 0:
					if ( array_key_exists( $key, $this->config->getDefaultUsed() ) ) {
						$attribs = array( 'class' => 'mw-zero-config-default' );
					} elseif ( array_key_exists( $key, $this->config->getSameAsDefault() ) ) {
						$attribs = array( 'class' => 'mw-zero-config-same' );
					} elseif ( array_key_exists( $key, $this->config->getUnknownFields() ) ) {
						$attribs = array( 'class' => 'mw-zero-config-unknown' );
					}
					break;
				case 1:
					if ( $parentKey === 'name' ) {
						$lang = Language::factory( $key );
						$fallbacks = $lang->getFallbackLanguages();
						if ( count( $fallbacks ) === 0 && $key !== 'en' ) {
							$fallbacks = array( 'en' );
						}
						if ( count( $fallbacks ) > 0 ) {
							$fbElems = array();
							foreach ( $fallbacks as $fb ) {
								$attr = null;
								if ( array_key_exists( $fb, $config ) ) {
									if ( $config[$fb] === $val ) {
										$attr = array( 'class' => 'mw-zero-config-same' );
									}
								} else {
									$attr = array( 'class' => 'mw-zero-config-unknown' );
								}
								$fbElems[] = Xml::element( 'span', $attr, $fb );
							}
							$td .= Xml::tags( 'td', array( 'class' => 'mw-zero-config-info' ), '&rarr; ' . implode( ' &rarr; ', $fbElems ) );
						}
					} elseif ( $parentKey === 'langNameOverrides' ) {
						$langName = Language::fetchLanguageName( $key );
						$original = $langName === $val ? Xml::span( $langName, 'mw-zero-config-same' ) : htmlspecialchars( $langName );
						$td .= Xml::tags( 'td', array( 'class' => 'mw-zero-config-info' ), $original );
					}
					break;
			}
			$rows[] = Xml::tags( 'tr', $attribs, $th . $td );
		}
		$res = Xml::tags( 'table', array( 'class' => 'mw-zero-config' ),
			Xml::tags( 'tbody', array(), join( "\n", $rows ) )
		);

		wfProfileOut( __METHOD__ );
		return $res;
	}

	private function renderBanners( $conf ) {
		global $wgConf;
		global $wgOut;

		wfProfileIn( __METHOD__ );
		$wgConf->loadFullData();

		// Merging keys preserves the order - first all the banners, then - name, etc
		$langs = array_merge(
			$conf['banner'], $conf['name'],
			array_flip( $conf['whitelistedLangs'] ),
			array_flip( $conf['showLangs'] ) );

		$configXcs = $wgOut->getTitle()->getRootText();

		$bannerSubdomain = !$conf['sites'] || in_array( 'm.wikipedia', $conf['sites'] ) ? 'm' : 'zero';

		$rows = array();

		$state = PageRenderingHooks::getState();
		foreach ( $langs as $langCode => $_ ) {
			$th = Xml::elementClean( 'th', null, $langCode );
			// render the banner without hiding it
			// Get localized "wikipedia" sitename. If null, will use default {{SITENAME}}
			$sitename = $wgConf->get( 'wgSitename', str_replace( '-', '_', $langCode ) . 'wiki' );

			$bannerInTheory = $state->renderBanner( $conf, $langCode, $sitename );
			$bannerInTheory = Xml::tags( 'td', null, $bannerInTheory ?: '???' );

			$iframe = Xml::tags( 'iframe', array(
					'src' => "//{$langCode}.{$bannerSubdomain}.wikipedia.org/wiki/MediaWiki:terms-use?X-CS={$configXcs}",
					'frameborder' => '0',
					'sandbox' => '', // Good idea to sandbox for new browsers!
					'scrolling' => 'no',
					'seamless' => 'seamless',
					'class' => 'mw-mf-iframe',
				),
				''
			);
			$bannerInPractice = Xml::tags( 'td', null, $iframe );
			$rows[] = Xml::tags( 'tr', null, $th . $bannerInTheory . $bannerInPractice );

		}
		$res = Xml::tags( 'table', array( 'class' => 'mw-zero-config-banners' ),
			Xml::tags( 'tbody', array(), join( "\n", $rows ) )
		);

		wfProfileOut( __METHOD__ );
		return $res;
	}
}
