<?php

namespace Extensions\ZeroRatedMobileAccess;

use Content;
use FormatJson;
use TextContentHandler;
use Title;

/**
 * JSON Zero Config content handler
 *
 * @file
 * @ingroup Extensions
 * @ingroup ZeroRatedMobileAccess
 *
 * @author Yuri Astrakhan <yurik@wikimedia.org>, based on Ori Livneh <ori@wikimedia.org> schema extension
 */
class ZeroConfigContentHandler extends TextContentHandler {

	/**
	 * @param string $modelId
	 */
	public function __construct( $modelId = 'JsonZeroConfig' ) {
		parent::__construct( $modelId, array( CONTENT_FORMAT_JSON ) );
	}

	/**
	 * Returns the content's text as-is.
	 *
	 * @param Content $content Content
	 * @param $format string|null
	 * @return mixed
	 */
	public function serializeContent( Content $content, $format = null ) {
		$this->checkFormat( $format );
		$conf = $content->getCarrierConfig();
		if ( !$conf->isValidJson() ) {
			// Invalid JSON - can't do anything with it
			return $content->getNativeData();
		}
		if ( !$conf->isError() ) {
			// There are no errors, normalize data (without defaults)
			$data = $conf->getData( false );
		} else {
			// JSON is valid, but the data has errors
			$data = $conf->getOriginalJson();
		}
		return FormatJson::encode( $data, true, FormatJson::ALL_OK );
	}

	/**
	 * Unserializes a JsonSchemaContent object.
	 *
	 * @param string $text Serialized form of the content
	 * @param null|string $format The format used for serialization
	 *
	 * @return Content the JsonSchemaContent object wrapping $text
	 */
	public function unserializeContent( $text, $format = null ) {
		$this->checkFormat( $format );
		static $lastText = null;
		static $lastRes = null;
		if ( $text !== $lastText ) {
			$lastText = $text;
			$lastRes = new ZeroConfigContent( $text );
		}
		return $lastRes;
	}

	/**
	 * Creates an empty JsonSchemaContent object.
	 *
	 * @return Content
	 */
	public function makeEmptyContent() {
		$default = <<<EOD
{
    "comment": "Full Company Name", // Optional
    "enabled": // Optional, true by default
    "sites": [
        // Optional, a list of values such as "zero.wikipedia" and "m.wikipedia"
        // Both values are used by default
    ],
    "name": {
        "en": "Company"
    },
    "banner": {
        "en": "Free {{SITENAME}} access by $1",
    },
    "bannerUrl": "http://wikipedia.org", // optional
    "showLangs": [
        "en"
        // required, which languages should be shown on the landing page.
        // If only one language is listed, the landing page will be redirected
        // to the Main Page, unless showZeroPage is set to override it.
    ],
    "whitelistedLangs": [
        // Required but can be an empty list = ALL, or a list of languages:  "en", "ru", ...
    ],
    "langNameOverrides": {
        "en": "Canadian" // Optional
    }
    "background": "#FFB612", // Optional, default = #E31230
    "foreground": "#000000", // Optional, default = #551011
    "fontSize": "0.9em", // Optional, skipped by default
    "bannerWarning": false, // Optional, default = true
    "showImages": false, // Optional, default = true
    "showZeroPage": false, // Optional, default is  count('showLangs') > 1 ? true : false
    "enableHttps": true, // Optional, default is false
    "ips": [
        // Neither of these two private CIDR ranges are legal
        "192.168.0.0/16",
        "10.0.0.0/8",
    ],
    "proxies": [
        "OPERA"
        // Optional. If not defined: no proxy zero-rating.
        // Otherwise, must be defined as array with elements
        // matching proxies with NS_ZERO configurations.
    ]
}
EOD;
		return new ZeroConfigContent( $default );
	}
}
