<?php

namespace Extensions\ZeroRatedMobileAccess;
use Language;
use UnlistedSpecialPage;
use Html;
use MobileContext;

/**
 * Default startup page for ZeroRatedMobileAccess extension that shows the list of available languages.
 *
 * @file
 * @ingroup Extensions
 */
class ZeroSpecialPage extends UnlistedSpecialPage {

	/**
	 * @var PageRendering
	 */
	private $state;

	/**
	 * Constructor — set up the SpecialPage:ZeroRatedMobileAccess page
	 */
	public function __construct() {
		parent::__construct( 'ZeroRatedMobileAccess' );
	}

	/**
	 * Show the special page
	 *
	 * @param null|string $par parameter passed to the special page or null
	 */
	public function execute( $par ) {
		$out = $this->getOutput();

		// todo: vary-headers should be removed from the mobile frontend
		$this->setHeaders();
		$out->addVaryHeader( 'X-CS' );
		$out->addVaryHeader( 'X-Subdomain' );
		$out->addVaryHeader( 'X-Forwarded-By' );
		// fixme: needs lower varnish caching time

		$this->state = PageRenderingHooks::getState( $this->getContext() );
		$req = $this->getRequest();
		if ( $req->getCheck( 'renderesibanner' ) ) {
			$this->echoEsiBanner( $this->state->getConfig(), $req->getCheck( 'wap' ), $req->getCheck( 'img' ) );
			return;
		}

		$redir = $this->state->getRedirectInfo();
		if ( array_key_exists( 'redirect', $redir ) ) {
			$out->redirect( $redir['redirect'], $redir['code'] );
			return;
		}

		$out->setPageTitle( null );
		if ( array_key_exists( 'warn', $redir ) ) {
			$html = $this->renderWarning( $redir );
			$out->addHTML( $html );
		} else {
			if ( array_key_exists( 'softredirect', $redir ) ) {
				$r = $redir['softredirect'];
				$rSafe = preg_match( '#^(https?://|/)#i', $r );
				$r = $rSafe ? wfUrlencode( $r ) : urlencode( $r );
				$output = '<a href="' . $r . '">' . htmlentities( $r ) . '</a>';
				$out->addHTML( $output );
			} else {
				$config = $this->state->getConfig();
				if ( $config !== null || $this->state->isZeroSubdomain() ) {
					$out->addModules( 'zero.special.scripts' );
					$out->addModuleStyles( 'zero.special.styles' );
					if ( $config !== null ) {
						$output = $this->renderSpecialPage( $config );
					} else {
						$output = $this->renderUnknownZeroPartnerPage();
					}
					$out->addHTML( $output );
				}
			}
		}
	}

	/**
	 * Render raw HTML snippet for ESI inclusion
	 * @param array $config
	 * @param bool $wap
	 * @param bool $isImage
	 */
	private function echoEsiBanner( $config, $wap, $isImage ) {
		$out = $this->getOutput();
		$out->disable();

		$request = $this->getRequest();
		$response = $request->response();

		$response->header( $out->getVaryHeader() );
		// $contentType = $this->getContentType();
		// $response->header( 'Content-type: ' . $contentType . '; charset=UTF-8' );

		// Both maxage and smaxage has to be between $min and $max
		$max = 900;
		$min = 5;
		$smaxage = min( $max, max( $min, abs( $request->getInt( 'smaxage', $max ) ) ) );
		$maxage = min( $max, max( $min, abs( $request->getInt( 'maxage', $max ) ) ) );

		$response->header( 'Cache-Control: public, s-maxage=' . $smaxage . ', max-age=' . $maxage );

		$banner = $this->state->renderBanner( $config, null, null, $wap, $isImage );

		echo $banner;
	}

	/**
	 * @param $config
	 * @return string
	 */
	private function renderSpecialPage( $config ) {
		$output = '';

		// CSS hack to get rid of the header that comes from the mobile frontend
		$output .= '<style type="text/css">.header {display: none}</style>';
		$output .= $this->msg( 'zero-landing-welcome' )->parseAsBlock() . Html::element( 'hr' );

		$languageNames = Language::fetchLanguageNames();
		$overrides = $config['langNameOverrides'];
		// Allow language name overrides, as requested by Orange Congo ('kg' => 'Kikongo')
		// Do it one at a time to keep the original order of the array
		foreach ( $overrides as $l => $n ) {
			$languageNames[$l] = $n;
		}

		foreach ( $config['showLangs'] as $languageCode ) {
			if ( !array_key_exists( $languageCode, $languageNames ) ) {
				continue;
			}
			$languageLink = Html::element( 'a',
				array( 'href' => $this->state->getStartPageUrl( $languageCode ) ),
				wfMessage( 'zero-rated-mobile-access-home-page-selection',
					ucfirst( $languageNames[$languageCode] ) )->inLanguage( $languageCode )
			);
			$output .= Html::rawElement( 'h3', null, $languageLink );
		}
		$output .= Html::element( 'hr' );
		$output .= wfMessage( 'zero-rated-mobile-access-home-page-selection-text' )->escaped();
		$output .= Html::openElement( 'form' );
		$currentTitle = $this->getPageTitle();
		$defaultUrl = MobileContext::singleton()->getMobileUrl( $currentTitle->getFullURL() );
		$output .= Html::element( 'input', array( 'type' => 'hidden',
			'name' => 'from',
			'value' => $currentTitle,
		) );
		// TODO: Remove unnecessary 'id' attributes
		$output .= Html::openElement( 'select',
			array( 'id' => 'languageselection',
				'onchange' => 'javascript:window.location = this.options[this.selectedIndex].value;',
				'class' => 'mw-zero-language',
				'name' => 'to',
			)
		);
		$output .= Html::element( 'option',
			array( 'value' => $defaultUrl ),
			wfMessage( 'zero-rated-mobile-access-language-selection' )->text()
		);

		$freeLangs = $config['whitelistedLangs'];
		foreach ( $languageNames as $languageCode => $languageName ) {
			$isFree = count( $freeLangs ) === 0 || in_array( $languageCode, $freeLangs );
			$url = $this->state->getStartPageUrl( $languageCode, !$isFree );
			if ( !$isFree ) {
				$url = $this->state->makeRedirect( $url );
			}
			$output .= Html::element( 'option', array( 'value' => $url ), $languageName );
		}
		$output .= Html::closeElement( 'select' );
		$output .= Html::element( 'input',
			array( 'type' => 'submit',
				'value' => wfMessage( 'go' )->text(),
				'class' => 'mw-zero-language',
			)
		);
		$output .= Html::closeElement( 'form' );
		return $output;
	}

	/**
	 * Create HTML warning of navigation to a paid resource
	 * @param array $redir that describes how redirect should be rendered
	 * @return string HTML
	 */
	private function renderWarning( $redir ) {
		// Show warning text (navigating away from the free site)
		$reject = Html::rawElement(
				'a',
				array(
					'class' => 'cancel inline zerobutton',
					'role' => 'button',
					'href' => $redir['from'] ),
				$this->msg( 'zero-rated-mobile-access-go-back' )->parse() ) . '';

		$accept = Html::rawElement(
				'a',
				array(
					'class' => 'button zerobutton',
					'role' => 'button',
					'href' => $redir['to'],
					'tabindex' => '0' ),
				$this->msg( 'zero-rated-mobile-access-accept' )->parse() );

		$lang = $this->getLanguage();
		$divOpen = '<div class="button-bar button-bar-centered buttonBar zerobar">';
		$divClose = '</div>';
		if ( $lang->isRTL() ) {
			$buttons = $divOpen . $accept . ' ' . $reject . $divClose;
		} else {
			$buttons = $divOpen . $reject . ' ' . $accept . $divClose;
		}
		$dir = $lang->getDir();
		$msg = $this->msg( $redir['warn'] === 'file'
			? 'zero-rated-mobile-access-file-auth'
			: 'zero-rated-mobile-access-charge-auth' )->parse();
		$html = <<<END
<div dir='$dir' class='header mw-mf-banner mw-mf-banner-undismissable'>$msg</div>
<div dir='$dir' class="zero-init-mw-viewPageTarget-toolbar-actions">$buttons</div>
END;
		return $html;
	}

	/**
	 * Render the warning text for the unknown IP when accessing zero.*
	 * @return string
	 */
	private function renderUnknownZeroPartnerPage() {

		// This page contains client IP and should not be cached
		$this->getOutput()->enableClientCache( false );
		$req = $this->getRequest();

		$info = $this->state->getWikiInfo();
		$host = implode( '.', $info ) . '.org';
		$bannerText = $this->msg( 'zero-rated-mobile-access-sorry', $host )->text() . '<br />';

		$ip = $req->getIP();
		$bannerText .= $this->msg( 'zero-rated-mobile-access-sorry-ip', $ip )->text() . '<br />';

		$url = false;
		$from = $req->getVal( 'from' );
		if ( $from ) {
			$urlParts = wfParseUrl( $from );
			if ( $urlParts && isset( $urlParts['host'] ) ) {
				// Check that this domain name is handled by local servers (disabled in testing)
				// Last two or three parts of the domain name must be found in the local vhost
				$hp = explode( '.', $urlParts['host'] );
				global $wgEnableZeroRatedMobileAccessTesting, $wgConf;
				if ( $wgEnableZeroRatedMobileAccessTesting ||
					$wgConf->isLocalVHost( implode( '.', array_slice( $hp, -2, 2 ) ) ) ||
					$wgConf->isLocalVHost( implode( '.', array_slice( $hp, -3, 3 ) ) )
				) {
					if ( $hp[1] === 'zero' ) {
						$hp[1] = 'm';
					} elseif ( $hp[0] === 'zero' ) {
						$hp[0] = 'm';
					}
					$urlParts['host'] = implode( '.', $hp );
					$url = wfUrlencode( wfAssembleUrl( $urlParts ) );
				}
			}
		}
		if ( !$url ) {
			$url = $this->state->getStartPageUrl( null, true ); // m. home page
		}
		$link = Html::element( 'a', array( 'href' => $url ), wfExpandUrl( $url ) );
		$bannerText .= $this->msg( 'zero-rated-mobile-access-sorry-goto', $link )->text();

		return Html::rawElement( 'div',
			array( 'class' => 'zrma-unknown' ),
			$bannerText );
	}
}
